"""
This module allows to create and use one of generic OAuth providers included in FastMCP.

FastMCP also supports a number of specific identity providers, e.g. Google or Auth0.
It offers a mechanism of configuring one of these providers using a set of defined
environment variables. The selection of the provider the MCP server is configured with
is defined in the environment variable FASTMCP_SERVER_AUTH.

For some reason FastMCP chose not to extend this mechanism on their generic providers.
This module does just that. It allows configuring one the three providers - JWTVerifier,
OAuthProxy, RemoteAuthProvider. To be precise, JWTVerifier can be configured using the
FastMCP environment variables, but because it is a part of the other two it is included
here as well. All optional parameters of these providers are supported, without giving
much thought on their usefulness in our use case.

If at any point in time FastMCP defines environment variables for the generic providers,
this module can be retired.

We deliberately use our own prefixes in the names of environment variables. If FastMCP
defines environment variables for the generic providers they can be used immediately
without waiting for this module to be removed.
"""

import csv
import os
from collections.abc import Callable
from dataclasses import dataclass
from functools import cache
from io import StringIO
from typing import Any

from fastmcp.server.auth import (
    AuthProvider,
    OAuthProxy,
    RemoteAuthProvider,
)
from fastmcp.server.auth.auth import TokenVerifier
from fastmcp.server.auth.providers.introspection import IntrospectionTokenVerifier
from fastmcp.server.auth.providers.jwt import JWTVerifier

ENV_PROVIDER_TYPE = "FASTMCP_SERVER_AUTH"


def str_to_list(s) -> list[str]:
    s = s.replace("\n", " ")
    reader = csv.reader(StringIO(s), skipinitialspace=True)
    fields = next(reader)
    return [fld.strip() for fld in fields if fld.strip()]


def str_to_str(s: str) -> str:
    fields = str_to_list(s)
    if fields:
        return fields[0]
    return ""


def str_to_bool(s) -> bool:
    s_lower = str_to_str(s).lower()
    if s_lower == "true":
        return True
    if s_lower == "false":
        return False
    raise ValueError(f"Invalid boolean parameter: {s}")


def str_to_int(s: str) -> int:
    return int(str_to_str(s))


def str_to_dict(s) -> dict[str, str]:
    fields = str_to_list(s)
    if (len(fields) % 2) != 0:
        raise ValueError(
            f"Invalid dictionary parameter: {s}. "
            "The dictionary must be provided as a plain list of keys and values."
        )
    return {fields[i]: fields[i + 1] for i in range(0, len(fields) - 1, 2)}


@dataclass
class AuthParameter:
    name: str
    conv: Callable[[str], Any] = lambda v: str_to_str(v)


@dataclass
class AuthProviderInfo:
    provider_type: type[AuthProvider]
    parameters: list[AuthParameter]


_generic_providers = [
    AuthProviderInfo(
        provider_type=JWTVerifier,
        parameters=[
            AuthParameter("public_key"),
            AuthParameter("jwks_uri"),
            AuthParameter("issuer"),
            AuthParameter("audience", str_to_list),
            AuthParameter("algorithm"),
            AuthParameter("required_scopes", str_to_list),
            AuthParameter("base_url"),
        ],
    ),
    AuthProviderInfo(
        provider_type=IntrospectionTokenVerifier,
        parameters=[
            AuthParameter("introspection_url"),
            AuthParameter("client_id"),
            AuthParameter("client_secret"),
            AuthParameter("timeout_seconds", str_to_int),
            AuthParameter("required_scopes", str_to_list),
            AuthParameter("base_url"),
        ],
    ),
    AuthProviderInfo(
        provider_type=RemoteAuthProvider,
        parameters=[
            AuthParameter("authorization_servers", str_to_list),
            AuthParameter("base_url"),
            AuthParameter("resource_name"),
            AuthParameter("resource_documentation"),
        ],
    ),
    AuthProviderInfo(
        provider_type=OAuthProxy,
        parameters=[
            AuthParameter("upstream_authorization_endpoint"),
            AuthParameter("upstream_token_endpoint"),
            AuthParameter("upstream_client_id"),
            AuthParameter("upstream_client_secret"),
            AuthParameter("upstream_revocation_endpoint"),
            AuthParameter("base_url"),
            AuthParameter("redirect_path"),
            AuthParameter("issuer_url"),
            AuthParameter("service_documentation_url"),
            AuthParameter("allowed_client_redirect_uris", str_to_list),
            AuthParameter("valid_scopes", str_to_list),
            AuthParameter("forward_pkce", str_to_bool),
            AuthParameter("token_endpoint_auth_method"),
            AuthParameter("extra_authorize_params", str_to_dict),
            AuthParameter("extra_token_params", str_to_dict),
        ],
    ),
]


def exa_provider_name(provider_type: type[AuthProvider]) -> str:
    return f"exa.{provider_type.__module__}.{provider_type.__qualname__}"


def exa_parameter_env_name(param: AuthParameter) -> str:
    # We don't use the class name in the environment variable names. In the future,
    # this can potentially create a name clash between the parameters of JWTVerifier
    # and either OAuthProxy or RemoteAuthProvider. This is very unlikely though,
    # and we will deal with if and when it happens.
    return f"EXA_AUTH_{param.name.upper()}"


@cache
def _get_provider_map() -> dict[str, AuthProviderInfo]:
    """
    Indexes all known providers by their names as they would apper in the
    FASTMCP_SERVER_AUTH envar.
    """
    return {
        exa_provider_name(provider.provider_type): provider
        for provider in _generic_providers
    }


@cache
def _get_verifier_map() -> dict[str, AuthProviderInfo]:
    """
    Indexes all known Token Verifiers by their names.
    """
    return {
        exa_provider_name(ver_type): ver_type
        for ver_type in [JWTVerifier, IntrospectionTokenVerifier]
    }


def create_auth_provider(
    provider_info: AuthProviderInfo, **extra_kwargs
) -> AuthProvider:
    kwargs = {
        param.name: param.conv(os.environ[exa_parameter_env_name(param)])
        for param in provider_info.parameters
        if exa_parameter_env_name(param) in os.environ
    }
    return provider_info.provider_type(**kwargs, **extra_kwargs)


def get_token_verifier(provider_name: str) -> tuple[TokenVerifier, str]:
    """
    Creates one of the known types of a Token Verifier. This can be either
    the requested Auth Provider itself, or a part of the requested Auth Provider.
    In the latter case, the function will create whatever Verifier it can create
    with the information given in the environment variables. Will raise ValueError
    if it can create none.
    Args:
        provider_name: requested Auth Provider.
    """

    # First check if the requested Auth provider is one of the Token Verifier types:
    provider_map = _get_provider_map()
    verifier_map = _get_verifier_map()
    provider_type = verifier_map.get(provider_name)
    if provider_type is not None:
        provider = create_auth_provider(provider_map[provider_name])
        return provider, provider_name

    # If the requested provider is not a Token Verifier than create a Token Verifier
    # that sufficient parameters are provided for.
    for ver_name, ver_type in verifier_map.items():
        try:
            verifier = create_auth_provider(provider_map[ver_name])
            return verifier, ver_name
        except ValueError:
            pass
    raise ValueError(
        "Insufficient parameters to create any of the supported "
        f"Token Verifier types: {verifier_map.keys()}"
    )


def get_auth_provider() -> AuthProvider | None:
    """
    Creates one of FastMCP generic OAuth2 providers, if the correspondent type name is
    set in the FASTMCP_SERVER_AUTH environment variable. Will raise ValueError if the
    requested provider type is unknown or no Token Verifier can be created.
    """
    provider_name = os.environ.get(ENV_PROVIDER_TYPE)
    if not provider_name:
        return None

    provider_map = _get_provider_map()
    if provider_name not in provider_map:
        raise ValueError(f"Generic OAuth provider {provider_name} is not supported ")

    verifier, verifier_name = get_token_verifier(provider_name)
    if provider_name == verifier_name:
        return verifier
    return create_auth_provider(provider_map[provider_name], token_verifier=verifier)


def get_auth_kwargs() -> dict[str, AuthProvider]:
    """
    A helper function that builds kwargs with an optional `auth` parameter,
    to be used in the MCP server constructor.
    """
    provider = get_auth_provider()
    if provider is None:
        return {}
    return {"auth": provider}



import threading
from collections import OrderedDict
from collections.abc import Callable
from typing import (
    Generic,
    TypeVar,
)

T = TypeVar("T")


class NamedObjectPool(Generic[T]):
    """
    A simple implementation of a thread-safe named object pool.

    There can be only one object in the pool with a given name. The pool has limited
    capacity. When it is exceeded, the least recently accessed object gets evicted.

    An object should not be shared between multiple threads. However, the class itself
    cannot enforce this. The consumer is supposed to return the object to the pool only
    when it has completely done with it.
    """

    def __init__(self, capacity: int, cleanup: Callable[[T], None] | None = None):
        self.capacity = capacity
        self.cleanup = cleanup
        self._available: OrderedDict[str, T] = OrderedDict()
        self._lock = threading.RLock()
        self._condition = threading.Condition(self._lock)

    def checkout(self, name: str) -> T | None:
        """
        Checks out an object by name, returns None if not available.
        """
        with self._condition:
            if name in self._available:
                return self._available.pop(name)
        return None

    def checkin(self, name: str, obj: T) -> None:
        """
        Returns an object to the pool.
        May need to evict the least recently added object from there to free the space.
        """
        with self._condition:
            if name in self._available:
                obj = self._available.pop(name)
                if self.cleanup is not None:
                    self.cleanup(obj)
            while len(self._available) >= self.capacity:
                _, obj = self._available.popitem(last=False)
                if self.cleanup is not None:
                    self.cleanup(obj)
            self._available[name] = obj



import ssl
from collections.abc import (
    Callable,
    Generator,
)
from contextlib import contextmanager
from pathlib import Path
from typing import (
    Any,
    ContextManager,
)

import exasol.saas.client.api_access as saas_api
import fastmcp.server.dependencies as fmcp_api
import pyexasol
import sqlglot.expressions as exp

from exasol.ai.mcp.server.named_object_pool import NamedObjectPool

ENV_DSN = "EXA_DSN"
""" Exasol DB server DSN """
ENV_USER = "EXA_USER"
""" The DB user name to be used by the MCP server """
ENV_PASSWORD = "EXA_PASSWORD"
""" The DB password for password authentication """
ENV_ACCESS_TOKEN = "EXA_ACCESS_TOKEN"
""" Bearer access token  """
ENV_REFRESH_TOKEN = "EXA_REFRESH_TOKEN"
""" Bearer refresh token  """
ENV_USERNAME_CLAIM = "EXA_USERNAME_CLAIM"
"""The name of the claim in the access token containing the DB username"""
ENV_POOL_SIZE = "EXA_POOL_SIZE"
"""The capacity of the connection pool"""
ENV_SAAS_HOST = "EXA_SAAS_HOST"
""" SaaS host, defaults to https://cloud.exasol.com/ """
ENV_SAAS_ACCOUNT_ID = "EXA_SAAS_ACCOUNT_ID"
""" SaaS account id """
ENV_SAAS_PAT = "EXA_SAAS_PAT"
""" SaaS PAT in case the server's own credentials are used to connect to SaaS DB """
ENV_SAAS_PAT_HEADER = "EXA_SAAS_PAT_HEADER"
""" Name of the header where the SaaS user's PAT is passed, e.g. x-api-key """
ENV_SAAS_DATABASE_ID = "EXA_SAAS_DATABASE_ID"
""" Name of the SaaS database id, if known """
ENV_SAAS_DATABASE_NAME = "EXA_SAAS_DATABASE_NAME"
""" Name of the SaaS database, if the id is unknown """
ENV_SSL_CERT_VALIDATION = "EXA_SSL_CERT_VALIDATION"
""" Verify other peers’ certificates (yes/no) """
ENV_SSL_TRUSTED_CA = "EXA_SSL_TRUSTED_CA"
""" Directory where Certification Authority (CA) certificates are stored, or a single CA file """
ENV_SSL_CLIENT_CERT = "EXA_SSL_CLIENT_CERT"
""" Own certificate file in PEM format """
ENV_SSL_PRIVATE_KEY = "EXA_SSL_PRIVATE_KEY"
""" Certificate's private key file """

DEFAULT_CONN_POOL_SIZE = 5
DEFAULT_SAAS_HOST = "https://cloud.exasol.com"


def local_env_complete(env: dict[str, Any]) -> bool:
    return all(v in env for v in [ENV_DSN, ENV_USER]) and any(
        v in env for v in [ENV_PASSWORD, ENV_ACCESS_TOKEN, ENV_REFRESH_TOKEN]
    )


def oidc_env_complete(env: dict[str, Any]) -> bool:
    return all(v in env for v in [ENV_DSN, ENV_USERNAME_CLAIM])


def saas_env_complete(env: dict[str, Any]) -> bool:
    return (
        (ENV_SAAS_ACCOUNT_ID in env)
        and any(v in env for v in [ENV_SAAS_PAT, ENV_SAAS_PAT_HEADER])
        and any(v in env for v in [ENV_SAAS_DATABASE_ID, ENV_SAAS_DATABASE_NAME])
    )


def _copy_kwargs(env: dict[str, Any], name_map: dict[str, str]) -> dict[str, Any]:
    return {val: env[key] for key, val in name_map.items() if key in env}


def get_local_kwargs(env: dict[str, Any]) -> dict[str, Any]:
    """
    Returns pyexasol.connect(...) arguments in case of On-Prem pre-configured
    server's credentials.
    """
    return _copy_kwargs(
        env,
        {
            ENV_DSN: "dsn",
            ENV_USER: "user",
            ENV_PASSWORD: "password",
            ENV_ACCESS_TOKEN: "access_token",
            ENV_REFRESH_TOKEN: "refresh_token",
        },
    )


def get_saas_kwargs(env: dict[str, Any]) -> dict[str, Any]:
    """
    Returns pyexasol.connect(...) arguments in case of SaaS backend.
    If PAT is not pre-configured, the function will attempt to extract it from the
    headers. The returned arguments is the result of an OpenAPI call that resolves
    the SaaS username and password from the provided account id and the database id
    or name.
    """
    saas_params = _copy_kwargs(
        env,
        {
            ENV_SAAS_HOST: "host",
            ENV_SAAS_ACCOUNT_ID: "account_id",
            ENV_SAAS_PAT: "pat",
            ENV_SAAS_DATABASE_ID: "database_id",
            ENV_SAAS_DATABASE_NAME: "database_name",
        },
    )
    if "host" not in saas_params:
        saas_params["host"] = DEFAULT_SAAS_HOST
    if "pat" not in saas_params:
        headers = fmcp_api.get_http_headers()
        if env[ENV_SAAS_PAT_HEADER] not in headers:
            raise RuntimeError(
                "The SAAS PAT is not specified and cannot be extracted from the call headers."
            )
        saas_params["pat"] = headers[env[ENV_SAAS_PAT_HEADER]]
    return saas_api.get_connection_params(**saas_params)


def optional_bool_from_env(env: dict[str, Any], var_name: str) -> bool | None:
    if var_name not in env:
        return None
    v = env[var_name].lower()
    if v in ["true", "yes", "y"]:
        return True
    elif v in ["false", "no", "n"]:
        return False
    else:
        raise ValueError(f"Unable to interpret {var_name} as a boolean value.")


def get_ssl_options(env: dict[str, Any]) -> dict[str, Any]:
    """
    Extracts SSL parameters from the provided configuration.
    Returns a dictionary in the websocket-client format
    (see https://websocket-client.readthedocs.io/en/latest/faq.html#what-else-can-i-do-with-sslopts)
    """
    sslopt: dict[str, object] = {}

    # Is server certificate validation required?
    certificate_validation = optional_bool_from_env(env, ENV_SSL_CERT_VALIDATION)
    if certificate_validation is not None:
        sslopt["cert_reqs"] = (
            ssl.CERT_REQUIRED if certificate_validation else ssl.CERT_NONE
        )

    # Is a bundle with trusted CAs provided?
    trusted_ca = env.get(ENV_SSL_TRUSTED_CA)
    if trusted_ca:
        trusted_ca_path = Path(trusted_ca)
        if trusted_ca_path.is_dir():
            sslopt["ca_cert_path"] = trusted_ca
        elif trusted_ca_path.is_file():
            sslopt["ca_certs"] = trusted_ca
        else:
            raise ValueError(f"Trusted CA location {trusted_ca} doesn't exist.")

    # Is client's own certificate provided?
    client_certificate = env.get(ENV_SSL_CLIENT_CERT)
    if client_certificate:
        if not Path(client_certificate).is_file():
            raise ValueError(f"Certificate file {client_certificate} doesn't exist.")
        sslopt["certfile"] = client_certificate
        private_key = env.get(ENV_SSL_PRIVATE_KEY)
        if private_key:
            if not Path(private_key).is_file():
                raise ValueError(f"Private key file {private_key} doesn't exist.")
            sslopt["keyfile"] = private_key

    return sslopt


def get_common_kwargs(env: dict[str, Any]) -> dict[str, Any]:
    common_kwargs = {
        "fetch_dict": True,
        "compression": True,
    }
    ssl_opt = get_ssl_options(env)
    if ssl_opt:
        common_kwargs["websocket_sslopt"] = ssl_opt
    return common_kwargs


def _create_connection_pool(
    env: dict[str:Any],
) -> NamedObjectPool[pyexasol.ExaConnection]:
    pool_size = int(env.get(ENV_POOL_SIZE, DEFAULT_CONN_POOL_SIZE))
    return NamedObjectPool(capacity=pool_size, cleanup=lambda conn: conn.close())


def _build_impersonate_query(user: str) -> str:
    # Currently, the IMPERSONATE query is not supported by SQLGlot.
    user_id = exp.Identifier(this=user, quoted=True)
    return f'IMPERSONATE {user_id.sql(dialect="exasol")}'


def get_oidc_user(username_claim: str | None) -> tuple[str | None, str | None]:
    token = fmcp_api.get_access_token()
    if token is None:
        return None, None
    return token.claims.get(username_claim), token.token


def get_connection_factory(
    env: dict[str:Any], **extra_kwargs
) -> Callable[[], ContextManager[pyexasol.ExaConnection]]:
    """
    Returns the pyexasol connection factory required by a DBConnection object.
    Authentication method will be inferred from the provided configuration
    parameters. Currently, the parameters come from environment variables.
    Going forward, the configuration parameters will be kept in the NBC secret store.

    For the On-Prem backend, the MCP server supports the same authentication methods as
    pyexasol. Currently, these are password and an OpenID token. To authenticate with
    the SaaS backend, the MCP server uses the Personal Access Token (PAT).

    The MCP server can be deployed in two ways: locally or as a remote http server.
    In the latter case the server works in the multiple user mode and its tools would
    normally be protected with OAuth2 authorization. The way to identify the user
    depends on the used backend.

    With an OnPrem database the MCP server can identify the user by looking at the
    claims in the access token. Most identity providers allow setting a custom claim
    or offer a choice of standard claims that can be used to store the DB username.
    The server needs to know the name of this claim.

    With a SaaS database, the PAT can be put in the call headers. The server needs to
    know the name of the header. In a way, the authentication is delegated to the SaaS
    database, the PAT itself being an access token. Therefore, a separate authorization
    of the MCP tools is optional.

    This gives us five different options for the database connection:
    *** On-Prem ***

    - A.
      The server is configured to use its own database credentials (username and either
      password or an OpenID token). No attempt is made to identify the actual user
      accessing the server tools. This works for both single and multiple user modes.
      The server tools may still be protected with OAuth2 authorization, but as far as
      the database connection is concerned this is irrelevant.
      The server's DB user must have the permission that is the least common denominator
      of the permissions of the users that are allowed to access the MCP server.

    - B.
      The server extracts the DB username, along with the token, from the MCP Auth
      context and uses that to open the connection. This option is suitable for
      multiuser mode, when the following two conditions are met:
      1. The users' authentication with the chosen identity provider is configured to
         add their DB usernames as a claim in the access token.
      2. The correspondent DB users are also authenticated using OpenID, with an access
         token (refresh token is currently not supported). The database verifies the
         token with the same identity provider as the MCP server. The subject, the DB
         user is identified with in the database, should, according to RFC 9068, match
         the subject field in the access token issued to this user.

    - C.
      The last option is a blend between the first two. It works in a multiuser mode,
      when the first of the above conditions is met but the second is not. The connection
      is opened using the pre-configured database credentials, as in the first option.
      But since the actual username can be identified, the connection impersonates this
      user. All subsequent queries are executed under this user's permissions. For this
      to work the server's user must have the "IMPERSONATE ANY USER" or "IMPERSONATION ON
      <user/role>" privilege.

      *** SaaS ***

      - D.
        The SaaS analogue of the option A. The server's database connection parameters
        include the SaaS host, account id, PAT and the database name or id.

      - E.
        The SaaS analogue of the option B, but much simpler. The PAT gets extracted from
        the headers and then used as in D.
    """

    # Validate the configuration. Here we only check the presence of a sufficient set
    # of environment variables to configure one of possible connection modes (Local,
    # OIDC or SaaS). Whether these variables are set correctly is another question. If
    # they are not, an exception will be raised in the factory. But we prefer it to be
    # raised here.
    if (
        (not local_env_complete(env))
        and (not oidc_env_complete(env))
        and (not saas_env_complete(env))
    ):
        raise ValueError("Insufficient database connection configuration")

    connection_pool = _create_connection_pool(env)

    @contextmanager
    def connection_factory() -> Generator[pyexasol.ExaConnection, None, None]:
        if saas_env_complete(env):
            conn_kwargs = get_saas_kwargs(env)
            user: str | None = None
        else:
            conn_kwargs = get_local_kwargs(env)
            user, token = get_oidc_user(env.get(ENV_USERNAME_CLAIM))
            if not local_env_complete(env):
                # If not using pre-configured server credentials then
                # authenticate with the token extracted from the MCP context.
                if user and token:
                    conn_kwargs["user"] = user
                    conn_kwargs["access_token"] = token
                else:
                    raise RuntimeError(
                        "Cannot extract user credentials from the MCP context, "
                        "and default credentials are not specified."
                    )
        user = user or conn_kwargs["user"]

        # Try to get the connection for the current user from the pool.
        connection = connection_pool.checkout(user)

        # Open a new one if needed.
        if (connection is None) or connection.is_closed:
            conn_kwargs.update(get_common_kwargs(env))
            conn_kwargs.update(extra_kwargs)
            connection = pyexasol.connect(**conn_kwargs)
            if user != conn_kwargs["user"]:
                # If the actual username is known, and it's different from the
                # username in the connection then impersonate the actual user.
                query = _build_impersonate_query(user)
                connection.execute(query)

        yield connection

        # Return the connection back to the pool, unless it has been closed.
        if not connection.is_closed:
            connection_pool.checkin(user, connection)

    return connection_factory
